import math
import imp
imp.load_source("util","ScriptEditor/util.py")  
from util import *
imp.load_source("boundaryUtil","ScriptEditor/boundaryUtil.py")  
from boundaryUtil import *

translator.install()
	
def computeBridgedL2DUtil1(L, sectionElements):
	tmp = [L]
	for sectionElement in sectionElements:
		if "l2d" in sectionElement:
			tmp[0] = tmp[0] - sectionElement["l2d"]
		else:
			l2dCount = 0
			while(True):
				l2dBind = "l2d_" + str(l2dCount)
				if not l2dBind in sectionElement:
					break
				if(l2dCount >= len(tmp)):
					tmp.append(L)
				tmp[l2dCount] = tmp[l2dCount] - sectionElement[l2dBind]				
				l2dCount = l2dCount + 1
	return tmp
							
def computeBridgedL2D(lComputations, lTableResults, flowTable, boundaryGroups, advancedResults):
	bridgedL2D = 0.0
	psi = [0.0]
	additionalPsi = list()

	if len(boundaryGroups) == 2:
		boundaryResults = advancedResults["boundary"]
		deltaT = computeDeltaT(boundaryResults, boundaryGroups)
		if deltaT != 0.0:
			bridgedL2D = computeFlow(boundaryResults) / deltaT
		else:
			bridgedL2D = 0.0
		
	elif len(boundaryGroups) == 3:
		boundaryResults = advancedResults["boundary"]
		boundariesAtSimulationTime = boundaryResults["boundaries_at_simulation_time"]
		intExtCouple = boundaryResults["int_ext_couple"]
		couplingsIDs = boundaryResults["monthly_couplings_ids"]
		thirdEnvHeated = False
		if "third_env_heated" in boundaryResults:
			thirdEnvHeated = boundaryResults["third_env_heated"]
		
		if intExtCouple >= 0 and intExtCouple < len(couplingsIDs):
			intExtCoupleIDs = couplingsIDs[intExtCouple]
			internalID = intExtCoupleIDs["internal"]
			externalID = intExtCoupleIDs["external"]

			internalGroupID = -1
			externalGroupID = -1
			otherGroupID = -1
			groupID = 0
			for boundaryGroup in boundaryGroups:
				for boundaryID in boundaryGroup:
					if boundaryID == internalID:
						internalGroupID = groupID
						break
					elif boundaryID == externalID:
						externalGroupID = groupID
						break
				groupID = groupID + 1
			for groupID in range(len(boundaryGroups)):
				if groupID != internalGroupID and groupID != externalGroupID:
					otherGroupID = groupID
					break

			if internalGroupID == -1 or externalGroupID == -1 or otherGroupID == -1:
				return bridgedL2D, psi, additionalPsi		
	
			internalGroup = boundaryGroups[internalGroupID]
			externalGroup = boundaryGroups[externalGroupID]
			otherGroup = boundaryGroups[otherGroupID]

			ieSectionElements = list()
			ioSectionElements = list()
			eoSectionElements = list()
			
			if "elements_1" in advancedResults:
				advancedResultsElements1 = advancedResults["elements_1"]
				for sectionElement in advancedResultsElements1.values():
					elementStartBoundary = sectionElement["start_boundary"]
					elementEndBoundary = sectionElement["end_boundary"]
					
					enfConf = 'auto'
					if 'env_conf' in sectionElement:
						enfConf = sectionElement['env_conf']
					
					if enfConf == 'auto':
						if(boundaryInGroup(elementStartBoundary, internalGroup) and boundaryInGroup(elementEndBoundary, externalGroup)) or\
							(boundaryInGroup(elementStartBoundary, externalGroup) and boundaryInGroup(elementEndBoundary, internalGroup)):
							ieSectionElements.append(sectionElement)
						
						if(boundaryInGroup(elementStartBoundary, internalGroup) and boundaryInGroup(elementEndBoundary, otherGroup)) or\
							(boundaryInGroup(elementStartBoundary, otherGroup) and boundaryInGroup(elementEndBoundary, internalGroup)):
							ioSectionElements.append(sectionElement)
						
						if(boundaryInGroup(elementStartBoundary, externalGroup) and boundaryInGroup(elementEndBoundary, otherGroup)) or\
							(boundaryInGroup(elementStartBoundary, otherGroup) and boundaryInGroup(elementEndBoundary, externalGroup)):
							eoSectionElements.append(sectionElement)
							
					elif enfConf == 'ie':
						ieSectionElements.append(sectionElement)
					elif enfConf == 'io':
						ioSectionElements.append(sectionElement)
					elif enfConf == 'eo':
						eoSectionElements.append(sectionElement)
						
				lIE = lTableResults[internalGroupID][str(externalGroupID)]
				lIO = lTableResults[internalGroupID][str(otherGroupID)]
				lEO = lTableResults[externalGroupID][str(otherGroupID)]
				
				tmp1 = computeBridgedL2DUtil1(lIE, ieSectionElements)
				tmp2 = computeBridgedL2DUtil1(lIO, ioSectionElements)
				tmp3 = computeBridgedL2DUtil1(lEO, eoSectionElements)
							
				internalT = boundariesAtSimulationTime[internalID]["temperature"]
				externalT = boundariesAtSimulationTime[externalID]["temperature"]
				otherT = boundariesAtSimulationTime[otherGroup[0]]["temperature"]
				deltaT = internalT - externalT
		
				if math.fabs(deltaT) > 0.0:
					for l2dCount in range(max(len(tmp1), len(tmp2))):
						if(l2dCount >= len(psi)):
							psi.append(0.0)
						if(l2dCount >= len(additionalPsi)):
							additionalPsi.append(dict())
							
						if l2dCount < len(tmp1):
							thisTmp1 = tmp1[l2dCount]
						else:
							thisTmp1 = tmp1[-1]
							
						if l2dCount < len(tmp2):
							thisTmp2 = tmp2[l2dCount]
						else:
							thisTmp2 = tmp2[-1]
						if l2dCount < len(tmp3):
							thisTmp3 = tmp3[l2dCount]
						else:
							thisTmp3 = tmp3[-1]

						psiIE = thisTmp1
						psiIO = thisTmp2 * (internalT - otherT) / deltaT
						psiEO = thisTmp3 * (otherT - externalT) / deltaT
						psi[l2dCount] = psiIE + psiIO
						additionalPsi[l2dCount]['ie'] = psiIE
						additionalPsi[l2dCount]['in'] = psiIO
						if thirdEnvHeated:
							psi[l2dCount] = psi[l2dCount] + psiEO
							additionalPsi[l2dCount]['en'] = psiEO

				fIE = flowTable[internalGroupID][str(externalGroupID)]
				fIO = flowTable[internalGroupID][str(otherGroupID)]

				#bridgedL2D will be set later as noBridgeL2D + psi
				bridgedL2D = 0.0

	return bridgedL2D, psi, additionalPsi
	
def computeBridgedL2DNoL(boundaryResults, deltaT):
	bridgedL2D = 0.0
	if math.fabs(deltaT) > 0.0:
		return computeFlow(boundaryResults) / deltaT
	else:
		return 0.0	
	
def computeNoBridgeL2D(deltaT,\
						advancedResults,\
						simulator,\
						currentGeometry,\
						currentFill,\
						currentBoundary,\
						Input,\
						Output,\
						OutputDescriptors):
	noBridgeL2D = [0.0]
	specialElementsInfo = simulator.getSpecialElementsInfo(currentGeometry)
	
	if not specialElementsInfo:
		return noBridgeL2D

	if specialElementsInfo[0]["element"]["__descriptor_name__"] == "__bridge_section_elements__":
		if "elements_1" in advancedResults:
			advancedResultsElements1 = advancedResults["elements_1"]
			for sectionElement in advancedResultsElements1.values():
				if "l2d" in sectionElement:
					noBridgeL2D[0] = noBridgeL2D[0] + sectionElement["l2d"]
				else:
					l2dCount = 0
					while(True):
						l2dBind = "l2d_" + str(l2dCount)
						if not l2dBind in sectionElement:
							break
						if(l2dCount >= len(noBridgeL2D)):
							noBridgeL2D.append(0.0)
						noBridgeL2D[l2dCount] = noBridgeL2D[l2dCount] + sectionElement[l2dBind]
						l2dCount = l2dCount + 1
					
	elif specialElementsInfo[0]["element"]["__descriptor_name__"] == "__bridge_areas__":

		noBridgeFlow = 0.0
		boundaryConditions = simulator.getBoundaryConditions(currentGeometry, currentFill, currentBoundary)
	
		specialElementsConfig = list()
		specialElementsConfig.append(1)
		Input["boundary_variant"] = "flows"
		simulationOutput, simulationOutputDesc = addSimulation(simulator,\
																currentGeometry,\
																currentFill,\
																currentBoundary,\
																specialElementsConfig,\
																boundaryConditions,\
																Input,\
																Output,\
																OutputDescriptors,\
																_("Bridge areas 1"))
		flow1 = computeFlow(simulationOutput["advanced_results"]["boundary"])
		noBridgeFlow = noBridgeFlow + flow1

		simulationOutputDesc.insert(0, createStringDescriptor("bridge_areas_id", "", "", -1, -1, False))
		simulationOutput["bridge_areas_id"] = "eqi"
		
		simulationOutputDesc.insert(0, createNumberDescriptor("l2d", "L2D", "W/mK", "%.4f", "l2d"))
		if deltaT != 0.0:
			simulationOutput["l2d"] = flow1 / deltaT
		else:
			simulationOutput["l2d"] = 0.0
			
		if "eqi_width" in simulationOutput:
			width = simulationOutput["eqi_width"]
			if width != 0.0 and deltaT != 0.0:
				simulationOutputDesc.insert(0, createNumberDescriptor("u", "Ueqi", "W/m²K", "%.4f", "u_eqi"))
				simulationOutput["u"] = flow1 / deltaT / width
			simulationOutput["width"] = width
		
		specialElementsConfig = list()
		specialElementsConfig.append(2)
		Input["boundary_variant"] = "flows"
		simulationOutput, simulationOutputDesc = addSimulation(simulator,\
																currentGeometry,\
																currentFill,\
																currentBoundary,\
																specialElementsConfig,\
																boundaryConditions,\
																Input,\
																Output,\
																OutputDescriptors,\
																_("Bridge areas 2"))
		flow2 = computeFlow(simulationOutput["advanced_results"]["boundary"])
		noBridgeFlow = noBridgeFlow + flow2

		simulationOutputDesc.insert(0, createStringDescriptor("bridge_areas_id", "", "", -1, -1, False))
		simulationOutput["bridge_areas_id"] = "eqo"
		
		simulationOutputDesc.insert(0, createNumberDescriptor("l2d", "L2D", "W/mK", "%.4f", "l2d"))
		if deltaT != 0.0:
			simulationOutput["l2d"] = flow2 / deltaT
		else:
			simulationOutput["l2d"] = 0.0
			
		if "eqo_width" in simulationOutput:
			width = simulationOutput["eqo_width"]
			if width != 0.0 and deltaT != 0.0:
				simulationOutputDesc.insert(0, createNumberDescriptor("u", "Ueqo", "W/m²K", "%.4f", "u_eqo"))
				simulationOutput["u"] = flow2 / deltaT / width
			simulationOutput["width"] = width
				
		if deltaT != 0.0:
			noBridgeL2D[0] = noBridgeFlow / math.fabs(deltaT)
		
	return noBridgeL2D

def computeNoBridgeL2D10077(boundaryGroups, \
							advancedResults, \
							panelResults, \
							bridgedL2D, \
							deltaT, \
							computeUfPsi, \
							computeUtjPsij):
	noBridgeL2D = [0.0]
	additionalResults = dict()
	if deltaT == 0.0:
		return noBridgeL2D, additionalResults

	panelAdvancedResults = panelResults["advanced_results"]
		
	bf = 0.0
	bp = 0.0
	bps = list()
	dp = list()
	flow = computeFlow(panelResults["advanced_results"]["boundary"])
	frameL2D = flow / deltaT
	
	numerator = frameL2D
	if "elements_1" in panelAdvancedResults:
		advancedResultsElements1 = advancedResults["elements_1"]
		panelAdvancedResultsElements1 = panelAdvancedResults["elements_1"]
		for sectionElementID, sectionElement in advancedResultsElements1.items():
			panelSectionElement = panelAdvancedResultsElements1[sectionElementID]

			if "length" in sectionElement:
				lengthBind = "length"
			else:
				lengthBind = "length_0"
			if "10077_insulation_panel" in sectionElement and sectionElement["10077_insulation_panel"]:
				if "l2d" in panelSectionElement:
					numerator = numerator - panelSectionElement["l2d"]
				else:
					numerator = numerator - panelSectionElement["l2d_0"]
					
				bp = bp + sectionElement[lengthBind]
				bps.append(sectionElement[lengthBind])
				dp.append(sectionElement["section_width"])
			else:					
				if not math.isnan(sectionElement[lengthBind]) :
					bf = bf + sectionElement[lengthBind]					
					
	glazingL2D = 0.0
	glazingL2Ds = list()
	glazingUs = list()
	
	if "elements_1" in advancedResults:
		advancedResultsElements1 = advancedResults["elements_1"]
		for sectionElementID, sectionElement in advancedResultsElements1.items():
			if "10077_insulation_panel" in sectionElement and sectionElement["10077_insulation_panel"]:				
				if "l2d" in sectionElement:
					glazingL2D = glazingL2D + sectionElement["l2d"]
					glazingL2Ds.append(sectionElement["l2d"])
				else:
					glazingL2D = glazingL2D + sectionElement["l2d_0"]
					glazingL2Ds.append(sectionElement["l2d_0"])
				glazingUs.append(sectionElement["u"])
					
	noBridgeL2D[0] = numerator + glazingL2D
	
	if bf != 0.0:
		uf = numerator / bf
	else:
		uf = 0.0
	
	if bf != 0.0:	
		if computeUtjPsij:
			if len(glazingL2Ds) == 1:
				utj = (bridgedL2D - glazingL2Ds[0]) / bf
				l1bf = bps[0] + bf / 2.0
				psij = bridgedL2D - glazingUs[0] * l1bf
				
				additionalResults["l1bf"] = l1bf
				additionalResults["utj"] = utj
				additionalResults["psij"] = psij
			if len(glazingL2Ds) == 2:
				utj = (bridgedL2D - glazingL2Ds[0] - glazingL2Ds[1]) / bf
				l1bf = bps[0] + bf / 2.0
				l2bf = bps[1] + bf / 2.0
				psij = bridgedL2D - glazingUs[0] * l1bf - glazingUs[1] * l2bf
				
				additionalResults["l1bf"] = l1bf
				additionalResults["l2bf"] = l2bf
				additionalResults["utj"] = utj
				additionalResults["psij"] = psij
	else:
		uf = 0.0
		
	additionalResults["flow"] = flow
	additionalResults["lf2d"] = frameL2D
	if computeUfPsi:
		additionalResults["uf"] = uf
	additionalResults["bf"] = bf
	additionalResults["bp"] = bp
	additionalResults["dp"] = dp
		
	return noBridgeL2D, additionalResults
	
def computeNoBridgeL1D_3D(advancedResults,\
						simulator,\
						currentGeometry):
	noBridgeL1D_3D = [0.0]
	specialElementsInfo = simulator.getSpecialElementsInfo(currentGeometry)
	
	if not specialElementsInfo:
		return noBridgeL1D_3D

	if specialElementsInfo[0]["element"]["__descriptor_name__"] == "__bridge_section_elements__":
		if "elements_1" in advancedResults:
			advancedResultsElements1 = advancedResults["elements_1"]
			for sectionElement in advancedResultsElements1.values():
				if "type" in sectionElement and sectionElement["type"] == "linear_3d":
					noBridgeL1D_3D[0] = noBridgeL1D_3D[0] + sectionElement["l"]
	return noBridgeL1D_3D	
	
def computeNoBridgeL2D_3D(advancedResults,\
						simulator,\
						currentGeometry):
	noBridgeL2D_3D = [0.0]
	specialElementsInfo = simulator.getSpecialElementsInfo(currentGeometry)
	
	if not specialElementsInfo:
		return noBridgeL2D_3D

	if specialElementsInfo[0]["element"]["__descriptor_name__"] == "__bridge_section_elements__":
		if "elements_1" in advancedResults:
			advancedResultsElements1 = advancedResults["elements_1"]
			for sectionElement in advancedResultsElements1.values():
				if "type" in sectionElement and sectionElement["type"] == "area_3d":
					noBridgeL2D_3D[0] = noBridgeL2D_3D[0] + sectionElement["l"]
	return noBridgeL2D_3D
					
def add10077UfResults(bridgedL2D,\
						additionalResults,\
						output,\
						outputDescriptors,\
						flowSimulation,\
						flowSimulationDescriptors,\
						panelSimulation10077,\
						panelSimulation10077Descriptors):
	if "flow" in additionalResults:
		output["flow"] = additionalResults["flow"]
		panelSimulation10077["flow"] = additionalResults["flow"]
	if "lf2d" in additionalResults:
		output["lf2d"] = additionalResults["lf2d"]
		panelSimulation10077["lf2d"] = additionalResults["lf2d"]
	flowSimulation["lf2d"] = bridgedL2D
	flowSimulation["flow"] = computeFlow(flowSimulation["advanced_results"]["boundary"])
		
	if "uf" in additionalResults:
		output["uf"] = additionalResults["uf"]
	if "bf" in additionalResults:
		output["bf"] = additionalResults["bf"]
	if "bp" in additionalResults:
		output["bp"] = additionalResults["bp"]

	hasPanels = "bp" in additionalResults and additionalResults["bp"] > 0.0
		
	if "dp" in additionalResults:
		dp = additionalResults["dp"]
	else:
		dp = list()
	
	for i in range(len(dp)):
		if(len(dp) > 1):
			bindSuffix = " " + str(i)
			tagSuffix = "_" + str(i)
		else:
			bindSuffix = ""
			tagSuffix = ""
			
		output["dp" + bindSuffix] = dp[i]
		outputDescriptors.insert(0, createNumberDescriptor("dp" + bindSuffix, _("Dp") + bindSuffix, "mm", "%.2f", "dp" + tagSuffix))
	
	if(hasPanels):
		outputDescriptors.insert(0, createNumberDescriptor("bp", "Bp", "mm", "%.2f", "bp"))
		
	outputDescriptors.insert(0, createNumberDescriptor("bf", "Bf" if hasPanels else "B", "mm", "%.2f", "bf"))
	
	if "uf" in additionalResults:
		outputDescriptors.insert(0, createNumberDescriptor("uf", "Uf" if hasPanels else "U", "W/m²K", "%.4f", "uf"))
		
	panelL2DName = _("Lf2D with panel")
	panelFlowName = _("Flow with panel")
	glazingL2DName = _("Lf2D with glazing")
	glazingFlowName = _("Flow with glazing")
	computingUtj = "utj" in additionalResults
		
	panelLf2dDescriptor = createNumberDescriptor("lf2d", panelL2DName if hasPanels else "L2D", "W/mK", "%.4f", "lf2d")
	panelFlowDescriptor = createNumberDescriptor("flow", panelFlowName if hasPanels else _("Flow"), "W/m", "%.4f", "flow")
	glazingLf2dDescriptor = createNumberDescriptor("lf2d", glazingL2DName, "W/mK", "%.4f", "lf2d")
	glazingFlowDescriptor = createNumberDescriptor("flow", glazingFlowName, "W/m", "%.4f", "flow")

	if len(panelSimulation10077Descriptors) != 0 or not computingUtj:
		outputDescriptors.insert(0, panelLf2dDescriptor)
		outputDescriptors.insert(0, panelFlowDescriptor)
	else:
		outputDescriptors.insert(0, glazingLf2dDescriptor)
		outputDescriptors.insert(0, glazingFlowDescriptor)
	
	if len(panelSimulation10077Descriptors) == 0 and not computingUtj:
		flowSimulationDescriptors.insert(0, panelLf2dDescriptor)	
		flowSimulationDescriptors.insert(0, panelFlowDescriptor)	
	else:
		flowSimulationDescriptors.insert(0, glazingLf2dDescriptor)	
		flowSimulationDescriptors.insert(0, glazingFlowDescriptor)	
		
		if len(panelSimulation10077Descriptors) != 0:
			panelSimulation10077Descriptors.insert(0, panelLf2dDescriptor)
			panelSimulation10077Descriptors.insert(0, panelFlowDescriptor)
			
	
	
	if computingUtj:
		output["l1bf"] = additionalResults["l1bf"]
		output["utj"] = additionalResults["utj"]
		output["psij"] = additionalResults["psij"]
		if "l2bf" in additionalResults:
			output["l2bf"] = additionalResults["l2bf"]
			outputDescriptors.insert(0, createNumberDescriptor("l2bf", "L2 + Bf/2", "mm", "%.2f", "l2bf", False))
		outputDescriptors.insert(0, createNumberDescriptor("l1bf", "L1 + Bf/2", "mm", "%.2f", "l1bf", False))
		outputDescriptors.insert(findDescriptor("uf", outputDescriptors) + 1, createNumberDescriptor("utj", _("UTj"), "W/m²K", "%.4f", "utj"))
		outputDescriptors.insert(0, createNumberDescriptor("psij", _("ψj"), "W/mK", "%.4f", "psij"))
	
def getSectionElementL2Ds(sectionElement):
	elementL2D = list()
	if "l2d" in sectionElement:
		elementL2D.append(sectionElement["l2d"])
	else:
		l2dCount = 0
		while(True):
			l2dBind = "l2d_" + str(l2dCount)			
			if not l2dBind in sectionElement:
				break
			elementL2D.append(sectionElement[l2dBind])
			l2dCount = l2dCount + 1
	return elementL2D
	
def addL2DResults(bridgedL2D,\
					noBridgeL2D,\
					psi,\
					additionalPsi,\
					input,\
					output,\
					outputDescriptors,\
					justPsi,\
					computeUfPsi):
	if len(psi) == 0 or psi[0] == 0.0:
		psi = []
		additionalPsi = []
		for i in range(len(noBridgeL2D)):
			psi.append(bridgedL2D - noBridgeL2D[i])
			additionalPsi.append(dict())
			
	# Remove elements psi from global psi
	advancedResults = input["advanced_results"]
	if "elements_1" in advancedResults:
		advancedResultsElements1 = advancedResults["elements_1"]
		for sectionElement in advancedResultsElements1.values():
			if "psi" in sectionElement:
				for i in range(len(psi)):
					psi[i] = psi[i] - sectionElement["psi"]
	
	paramCount = 0

	for i in range(len(noBridgeL2D)):
		if len(noBridgeL2D) > 1:
			namePostfix = " " + str(i + 1)
			bindPostfix = "_" + str(i)
		else:
			namePostfix = ""
			bindPostfix = ""

		if math.fabs(psi[i]) > 0.000001 and noBridgeL2D[i] != 0.0 and computeUfPsi:
			outputDescriptors.insert(paramCount, createNumberDescriptor("psi" + bindPostfix,\
																		"ψ" + namePostfix,\
																		"W/mK",\
																		"%.4f",\
																		"psi" + bindPostfix))
			paramCount = paramCount + 1
			for additionalPsiKey, additionalPsiValue in additionalPsi[i].items():
				outputDescriptors.insert(paramCount, createNumberDescriptor("psi" + bindPostfix + additionalPsiKey,\
																			"ψ" + namePostfix + additionalPsiKey,\
																			"W/mK",\
																			"%.4f",\
																			"psi" + bindPostfix + additionalPsiKey))
				paramCount = paramCount + 1
		
		output["psi" + bindPostfix] = psi[i]
		for additionalPsiKey, additionalPsiValue in additionalPsi[i].items():
			output["psi" + bindPostfix + additionalPsiKey] = additionalPsiValue
		output["l2d_no_bridge" + bindPostfix] = noBridgeL2D[i]
		
	if bridgedL2D != 0.0 and not justPsi:
		output["l2d_bridged"] = bridgedL2D
		outputDescriptors.insert(paramCount, createNumberDescriptor("l2d_bridged", _("Bridged L2D"), "W/mK", "%.4f", "l2d_bridged"))
		paramCount = paramCount + 1	
		
	for i in range(len(noBridgeL2D)):
		if len(noBridgeL2D) > 1:
			namePostfix = " " + str(i + 1)
			bindPostfix = "_" + str(i)
		else:
			namePostfix = ""
			bindPostfix = ""
			
		if noBridgeL2D[i] != 0.0 and not justPsi:
			outputDescriptors.insert(paramCount, createNumberDescriptor("l2d_no_bridge" + bindPostfix, _("No bridge L1D") + namePostfix, "W/mK", "%.4f", "l2d_no_bridge" + bindPostfix))
			paramCount = paramCount + 1		
		
	if "elements_1" in advancedResults:
		processedSectionElements = list()
		advancedResultsElements1 = advancedResults["elements_1"]
		for sectionElement in advancedResultsElements1.values():
			processedSectionElements.append(False)
					
		if "u_factors" in input:
			uFactorsGroups = dict()
			for uFactorKey, uFactor in input["u_factors"].items():
				
				if not isinstance(uFactor, dict):
					continue
				group = uFactor["group"]
				if group == "":
					continue
					
				groupL1D = list()					
				for sectionElement in advancedResultsElements1.values():
					for touchedBoundary in uFactor["touched_boundaries"]:
						sectionElementCount = 0
						for sectionElement in advancedResultsElements1.values():
							if not processedSectionElements[sectionElementCount]:
								elementStartBoundary = sectionElement["start_boundary"]
								elementEndBoundary = sectionElement["end_boundary"]
								if elementStartBoundary == touchedBoundary or elementEndBoundary == touchedBoundary:
									processedSectionElements[sectionElementCount] = True
									elementL2Ds = getSectionElementL2Ds(sectionElement)
									l2DCount = 0
									for elementL2D in elementL2Ds:
										if l2DCount >= len(groupL1D):
											groupL1D.append(0.0)
										groupL1D[l2DCount] = groupL1D[l2DCount] + elementL2D
										l2DCount = l2DCount + 1
							sectionElementCount = sectionElementCount + 1
				
				if not group in uFactorsGroups:
					uFactorsGroups[group] = dict()
					uFactorsGroups[group]["l1d"] = list()
								
				while len(uFactorsGroups[group]["l1d"]) < len(groupL1D):
					uFactorsGroups[group]["l1d"].append(0.0)
				
				for i in range(len(groupL1D)):
					uFactorsGroups[group]["l1d"][i] = uFactorsGroups[group]["l1d"][i] + groupL1D[i]
				
				l2D = uFactor["l2d"]
				if not "l2d" in uFactorsGroups[group]:
					uFactorsGroups[group]["l2d"] = 0.0	
				uFactorsGroups[group]["l2d"] = uFactorsGroups[group]["l2d"] + l2D
				
			groupCount = 0
			for uFactorsGroup, uFactorValue in uFactorsGroups.items():
				for i in range(len(uFactorValue["l1d"])):
					l1D = uFactorValue["l1d"][i]					
					if len(uFactorValue["l1d"]) > 1:
						namePostfix = " " + str(i + 1)
						bindPostfix = "_" + str(i)
					else:
						namePostfix = ""
						bindPostfix = ""
			
					bind = uFactorsGroup + "_psi" + bindPostfix
					outputDescriptors.insert(paramCount, createNumberDescriptor(bind, uFactorsGroup + " ψ" + namePostfix, "W/mK", "%.4f", "psi_" + str(groupCount) + bindPostfix))
					
					output[bind] = uFactorValue["l2d"] - l1D
					paramCount = paramCount + 1
				groupCount = groupCount + 1		
	
def addL1DResults_3D(noBridgeL1D_3D, output, outputDescriptors):
	paramCount = 0
	for i in range(len(noBridgeL1D_3D)):
		if len(noBridgeL1D_3D) > 1:
			namePostfix = " " + str(i + 1)
			bindPostfix = "_" + str(i)
		else:
			namePostfix = ""
			bindPostfix = ""
			
		bind = "l1d_3d_no_bridge"			
		output[bind] = noBridgeL1D_3D[i]
		
		if noBridgeL1D_3D[i] != 0.0:
			outputDescriptors.insert(paramCount, createNumberDescriptor(bind, _("Σψ*l") + namePostfix, "W/K", "%.4f", bind))
			paramCount = paramCount + 1	
	
def addL2DResults_3D(noBridgeL2D_3D, output, outputDescriptors):
	paramCount = 0
	for i in range(len(noBridgeL2D_3D)):
		if len(noBridgeL2D_3D) > 1:
			namePostfix = " " + str(i + 1)
			bindPostfix = "_" + str(i)
		else:
			namePostfix = ""
			bindPostfix = ""
			
		bind = "l2d_3d_no_bridge"			
		output[bind] = noBridgeL2D_3D[i]
		
		if noBridgeL2D_3D[i] != 0.0:
			outputDescriptors.insert(paramCount, createNumberDescriptor(bind, _("ΣU*A") + namePostfix, "W/K", "%.4f", bind))
			paramCount = paramCount + 1

def addPSIResults_3D(L3D,\
					noBridgeL1D_3D,\
					noBridgeL2D_3D,\
					output,\
					outputDescriptors):

	for i in range(len(noBridgeL1D_3D)):
		if len(noBridgeL2D_3D) > 1:
			namePostfix = " " + str(i + 1)
			bindPostfix = "_" + str(i)
		else:
			namePostfix = ""
			bindPostfix = ""
		
		psi = L3D - noBridgeL1D_3D[i] - noBridgeL2D_3D[i]
		if psi != 0.0:
			outputDescriptors.insert(0, createNumberDescriptor("psi_3d" + bindPostfix,\
																"χ" + namePostfix,\
																"W/K",\
																"%.4f",\
																"psi_3d" + bindPostfix))
			output["psi_3d" + bindPostfix] = psi
		
def addL3DResults_3D(L3D, output, outputDescriptors):
	if L3D != 0.0:
		output["l3d"] = L3D
		outputDescriptors.insert(0, createNumberDescriptor("l3d", _("L3D"), "W/K", "%.4f", "l3d"))		
		
def addUResults(bridgedL2D, input, output, outputDescriptors):
	if "width" in input:
		width = input["width"]
	else:
		width = 0.0
	if width != 0.0:
		U = bridgedL2D / width
		outputDescriptors.insert(0, createNumberDescriptor("u", _("Uav"), "W/m²K", "%.4f", "u"))
		output["u"] = U
	else:
		output["u"] = 0.0
			
def addFlowResults(bridgedL2D, noBridgeL2D, psi, deltaT, output, outputDescriptors):
	if math.fabs(deltaT) > 0.0001:
		bridgedAdded = False
		for i in range(len(noBridgeL2D)):
			if len(noBridgeL2D) > 1:
				namePostfix = " " + str(i + 1)
				bindPostfix = "_" + str(i)
			else:
				namePostfix = ""
				bindPostfix = ""						

			if noBridgeL2D[i] > 0.0:
				outputDescriptors.insert(findDescriptor("l2d_no_bridge" + bindPostfix, outputDescriptors) + 1,\
										createNumberDescriptor("flow_no_bridge" + bindPostfix,\
										_("Φ, no bridge") + namePostfix,\
										"W/m",\
										"%.4f",\
										"flow_no_bridge" + bindPostfix))

			output["flow_no_bridge" + bindPostfix] = noBridgeL2D[i] * deltaT

		if not bridgedAdded and bridgedL2D != 0.0:
			output["flow_bridged"] = bridgedL2D * deltaT
			outputDescriptors.insert(findDescriptor("l2d_bridged", outputDescriptors) + 1,\
									createNumberDescriptor("flow_bridged",\
									_("Av. Φ, bridged"),\
									"W/m",\
									"%.4f",\
									"flow_bridged"))
		
		deltaTDescriptorPos = findDescriptor("flow_bridged", outputDescriptors)
		if deltaTDescriptorPos > 0:
			deltaTDescriptorPos = deltaTDescriptorPos - 1
		else:
			deltaTDescriptorPos = findDescriptor("advanced_results", outputDescriptors)
		outputDescriptors.insert(deltaTDescriptorPos,\
								createNumberDescriptor("delta_t", _("ΔT"),\
								"°C",\
								"%.2f",\
								"delta_t"))
		output["delta_t"] = deltaT
